"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from twitchio.user import PartialUser
from twitchio.utils import parse_timestamp


if TYPE_CHECKING:
    import datetime

    from twitchio.http import HTTPClient
    from twitchio.types_.responses import (
        AutomodSettingsResponseData,
        BannedUsersResponseData,
        BanUserResponseData,
        BlockedTermsResponseData,
        CheckAutomodStatusResponseData,
        ResolveUnbanRequestsResponseData,
        ShieldModeStatusResponseData,
        UnbanRequestsResponseData,
        WarnChatUserResponseData,
    )

__all__ = (
    "AutoModStatus",
    "AutomodCheckMessage",
    "AutomodSettings",
    "Ban",
    "BannedUser",
    "BlockedTerm",
    "ShieldModeStatus",
    "Timeout",
    "UnbanRequest",
    "Warning",
)


class AutoModStatus:
    """Represents the Automod status of a message.

    Attributes
    -----------
    msg_id: str
        A caller-defined ID passed in the request.
    permitted: bool
        A Boolean value that indicates whether Twitch would approve the message for chat or hold it for moderator review or block it from chat.
        Is True if Twitch would approve the message; otherwise, False if Twitch would hold the message for moderator review or block it from chat.
    """

    __slots__ = ("msg_id", "permitted")

    def __init__(self, data: CheckAutomodStatusResponseData) -> None:
        self.msg_id: str = data["msg_id"]
        self.permitted: bool = data["is_permitted"]

    def __repr__(self) -> str:
        return f"<AutoModStatus msg_id={self.msg_id} permitted={self.permitted}>"


class AutomodCheckMessage:
    """Represents the message to check with automod

    Attributes
    -----------
    id: str
        Developer-generated identifier for mapping messages to results.
    text: str
        Message text.
    """

    __slots__ = ("id", "text")

    def __init__(self, id: str, text: str) -> None:
        self.id = id
        self.text = text

    def _to_dict(self) -> dict[str, str]:
        return {"msg_id": self.id, "msg_text": self.text}

    def __repr__(self) -> str:
        return f"<AutomodCheckMessage id={self.id} text={self.text}>"


class AutomodSettings:
    """Represents the AutoModSettings of a broadcaster's chat room.

    Attributes
    -----------
    broadcaster: PartialUser
        The PartialUser instance representing the broadcaster.
    moderator: PartialUser
        The PartialUser instance representing the moderator.
    overall_level: int | None
        The overall moderation level, which could be None if the broadcaster has set one or more of the individual settings.
    disability: int
        The Automod level for discrimination against disability.
    aggression: int
        The Automod level for hostility involving aggression.
    sexuality_sex_or_gender : int
        The AutoMod level for discrimination based on sexuality, sex, or gender.
    misogyny: int
        The Automod level for discrimination against women.
    bullying: int
        The Automod level for hostility involving name calling or insults.
    swearing: int
        The Automod level for profanity.
    race_ethnicity_or_religion: int
        The Automod level for racial discrimination.
    sex_based_terms: int
        The Automod level for sexual content.
    """

    __slots__ = (
        "aggression",
        "broadcaster",
        "bullying",
        "disability",
        "misogyny",
        "moderator",
        "overall_level",
        "race_ethnicity_or_religion",
        "sex_based_terms",
        "sexuality_sex_or_gender",
        "swearing",
    )

    def __init__(self, data: AutomodSettingsResponseData, *, http: HTTPClient) -> None:
        self.broadcaster: PartialUser = PartialUser(data["broadcaster_id"], None, http=http)
        self.moderator: PartialUser = PartialUser(data["moderator_id"], None, http=http)
        self.overall_level: int | None = int(data["overall_level"]) if data["overall_level"] else None
        self.disability: int = data["disability"]
        self.aggression: int = data["aggression"]
        self.sexuality_sex_or_gender: int = data["sexuality_sex_or_gender"]
        self.misogyny: int = data["misogyny"]
        self.bullying: int = data["bullying"]
        self.swearing: int = data["swearing"]
        self.race_ethnicity_or_religion: int = data["race_ethnicity_or_religion"]
        self.sex_based_terms: int = data["sex_based_terms"]

    def __repr__(self) -> str:
        return (
            f"<AutomodSettings broadcaster={self.broadcaster} moderator={self.moderator} overall_level={self.overall_level}>"
        )

    def to_dict(self, use_ids: bool = False) -> dict[str, str | int | None]:
        """Returns the AutomodSettings as a dictionary. This is the equivalent of the raw payload returned by Twitch.

        Parameters
        ----------
        use_ids : bool
            Whether to include the broadcaster and moderator IDs in the dictionary.

        Returns
        -------
        dict[str, str | int | None]
            Dictionary data type of AutomodSettings
        """
        result: dict[str, str | int | None] = {}
        user_related = {"broadcaster", "moderator"}

        if use_ids:
            for user_attr in user_related:
                user = getattr(self, user_attr, None)
                result[f"{user_attr}_id"] = getattr(user, "id", None) if user else None

        for attribute in self.__slots__:
            if attribute in user_related and not use_ids:
                continue
            attr_value = getattr(self, attribute, None)
            if attr_value is not None and hasattr(attr_value, "to_dict"):
                result[attribute] = attr_value.to_dict()
            elif attribute not in user_related:
                result[attribute] = attr_value

        return result


class BannedUser:
    """Represents a BannedUser.

    Attributes
    ----------
    user: PartialUser
        The user banned or timed out.
    moderator: PartialUser
        The moderator who banned or put the user in timeout.
    expires_at: datetime.datetime | None
        Datetime of when the timeout will end. This is None if permanently banned.
    created_at: datetime.datetime
        Datetime of when the user was banned.
    reason: str
        The reason the user was banned or put in a timeout if the moderator provided one.
    """

    __slots__ = ("created_at", "expires_at", "moderator", "reason", "user")

    def __init__(self, data: BannedUsersResponseData, *, http: HTTPClient) -> None:
        self.user: PartialUser = PartialUser(data["user_id"], data["user_login"], data["user_name"], http=http)
        self.moderator: PartialUser = PartialUser(
            data["moderator_id"], data["moderator_login"], data["moderator_name"], http=http
        )
        self.expires_at: datetime.datetime | None = parse_timestamp(data["expires_at"]) if data["expires_at"] else None
        self.created_at: datetime.datetime = parse_timestamp(data["created_at"])
        self.reason: str = data["reason"]

    def __repr__(self) -> str:
        return f"<BannedUser user={self.user} created_at={self.created_at} expires_at={self.expires_at}>"


class Ban:
    """Represents a Ban.

    Attributes
    ----------
    broadcaster: PartialUser
        The broadcaster whose chat room the user was banned from chatting in.
    user: PartialUser
        The user banned or timed out.
    moderator: PartialUser
        The moderator who banned or put the user in timeout.
    created_at: datetime.datetime
        Datetime of when the user was banned.
    """

    __slots__ = ("broadcaster", "created_at", "end_time", "moderator", "user")

    def __init__(self, data: BanUserResponseData, *, http: HTTPClient) -> None:
        self.broadcaster: PartialUser = PartialUser(data["broadcaster_id"], None, http=http)
        self.user: PartialUser = PartialUser(data["user_id"], None, http=http)
        self.moderator: PartialUser = PartialUser(data["moderator_id"], None, http=http)
        self.created_at: datetime.datetime = parse_timestamp(data["created_at"])

    def __repr__(self) -> str:
        return f"<Ban broadcaster={self.broadcaster} user={self.user} created_at={self.created_at}>"


class Timeout:
    """Represents a Timeout.

    Attributes
    ----------
    broadcaster: PartialUser
        The broadcaster whose chat room the user was timed out from chatting in.
    user: PartialUser
        The user timed out.
    moderator: PartialUser
        The moderator who put the user in timeout.
    end_time: datetime.datetime
        Datetime of when the timeout will end.
    created_at: datetime.datetime
        Datetime of when the user was timed out.
    """

    __slots__ = ("broadcaster", "created_at", "end_time", "moderator", "user")

    def __init__(self, data: BanUserResponseData, *, http: HTTPClient) -> None:
        self.broadcaster: PartialUser = PartialUser(data["broadcaster_id"], None, http=http)
        self.user: PartialUser = PartialUser(data["user_id"], None, http=http)
        self.moderator: PartialUser = PartialUser(data["moderator_id"], None, http=http)
        self.end_time: datetime.datetime | None = parse_timestamp(data["end_time"]) if data["end_time"] else None
        self.created_at: datetime.datetime = parse_timestamp(data["created_at"])

    def __repr__(self) -> str:
        return f"<Timeout broadcaster={self.broadcaster} user={self.user} created_at={self.created_at} end_time={self.end_time}>"


class UnbanRequest:
    """Represents an unban request.

    Attributes
    ----------
    id: str
        Unban request ID.
    broadcaster: PartialUser
        The broadcaster whose channel is receiving the unban request.
    moderator: PartialUser
        The moderator who approved/denied the request.
    user: PartialUser
        The requestor who is asking for an unban.
    text: str
        Text of the request from requesting user.
    status: typing.Literal["pending", "approved", "denied", "acknowledged", "canceled"]
        Status of the request. One of: `pending`, `approved`, `denied`, `acknowledged`, `canceled`
    created_at: datetime.datetime
        Datetime of when the unban request was created.
    resolved_at: datetime.datetime | None
        Datetime of when moderator/broadcaster approved or denied the request.
    resolution_text: str | None
        Text input by the resolver (moderator) of the unban request.
    """

    __slots__ = (
        "broadcaster",
        "created_at",
        "id",
        "moderator",
        "resolution_text",
        "resolved_at",
        "status",
        "text",
        "user",
    )

    def __init__(self, data: UnbanRequestsResponseData | ResolveUnbanRequestsResponseData, *, http: HTTPClient) -> None:
        self.id: str = data["id"]
        self.broadcaster: PartialUser = PartialUser(
            data["broadcaster_id"], data["broadcaster_login"], data["broadcaster_name"], http=http
        )
        self.moderator: PartialUser = PartialUser(
            data["moderator_id"], data["moderator_login"], data["moderator_name"], http=http
        )
        self.user: PartialUser = PartialUser(data["user_id"], data["user_login"], http=http)
        self.text: str = data["text"]
        self.status: Literal["pending", "approved", "denied", "acknowledged", "canceled"] = data["status"]
        self.created_at: datetime.datetime = parse_timestamp(data["created_at"])
        self.resolved_at: datetime.datetime | None = parse_timestamp(data["resolved_at"]) if data["resolved_at"] else None
        self.resolution_text: str | None = data["resolution_text"] or None

    def __repr__(self) -> str:
        return f"<UnbanRequest id={self.id} broadcaster={self.broadcaster} user={self.user} created_at={self.created_at}>"


class BlockedTerm:
    """
    Represents a blocked term.

    Attributes
    ----------
    id: str
        An ID that identifies this blocked term.
    broadcaster: PartialUser
        The broadcaster that owns the list of blocked terms.
    moderator: PartialUser
        The moderator that blocked the word or phrase from being used in the broadcaster's chat room.
    text: str
        The blocked word or phrase.
    created_at: datetime.datetime
        Datetime that the term was blocked.
    updated_at: datetime.datetime
        Datetime that the term was updated. When the term is added, this timestamp is the same as `created_at`. The timestamp changes as `AutoMod` continues to deny the term.
    expires_at: datetime.datetime | None
        Datetime that the blocked term is set to expire. After the block expires, users may use the term in the broadcaster's chat room.
        Is None if the term was added manually or was permanently blocked by AutoMod.
    """

    __slots__ = ("broadcaster", "created_at", "expires_at", "id", "moderator", "text", "updated_at")

    def __init__(self, data: BlockedTermsResponseData, http: HTTPClient) -> None:
        self.id: str = data["id"]
        self.broadcaster: PartialUser = PartialUser(data["broadcaster_id"], None, http=http)
        self.moderator: PartialUser = PartialUser(data["moderator_id"], None, http=http)
        self.text: str = data["text"]
        self.created_at: datetime.datetime = parse_timestamp(data["created_at"])
        self.updated_at: datetime.datetime = parse_timestamp(data["updated_at"])
        self.expires_at: datetime.datetime | None = parse_timestamp(data["expires_at"]) if data["expires_at"] else None


class ShieldModeStatus:
    """Represents a shield mode status..

    Attributes
    ----------
    active: bool
        Whether Shield Mode is active. Is true if Shield Mode is active;
    moderator: PartialUser
        The moderator that last activated Shield Mode.
    last_activated_at: datetime.datetime
        When Shield Mode was last activated.
    """

    __slots__ = ("active", "last_activated_at", "moderator")

    def __init__(self, data: ShieldModeStatusResponseData, *, http: HTTPClient) -> None:
        self.active: bool = bool(data["is_active"])
        self.moderator: PartialUser = PartialUser(
            data["moderator_id"], data["moderator_login"], data["moderator_name"], http=http
        )
        self.last_activated_at: datetime.datetime = parse_timestamp(data["last_activated_at"])

    def __repr__(self) -> str:
        return (
            f"<ShieldModeStatus active={self.active} moderator={self.moderator} last_activated_at={self.last_activated_at}>"
        )


class Warning:
    """Represents a warning to a user.

    Attributes
    ----------
    broadcaster: PartialUser
        The channel in which the warning will take effect.
    user: PartialUser
        The warned user.
    moderator: PartialUser
        The moderator who applied the warning.
    reason: str
        The reason provided for warning.
    """

    __slots__ = ("broadcaster", "moderator", "reason", "user")

    def __init__(self, data: WarnChatUserResponseData, *, http: HTTPClient) -> None:
        self.broadcaster: PartialUser = PartialUser(data["broadcaster_id"], None, http=http)
        self.user: PartialUser = PartialUser(data["user_id"], None, http=http)
        self.moderator: PartialUser = PartialUser(data["moderator_id"], None, http=http)
        self.reason: str = data["reason"]

    def __repr__(self) -> str:
        return f"<Warning broadcaster={self.broadcaster} user={self.user} moderator={self.moderator} reason={self.reason}>"



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from twitchio.types_.responses import (
    CheckUserSubscriptionResponseData,
)
from twitchio.user import PartialUser


if TYPE_CHECKING:
    from twitchio.http import HTTPAsyncIterator, HTTPClient
    from twitchio.types_.responses import (
        BroadcasterSubscriptionsResponse,
        BroadcasterSubscriptionsResponseData,
        CheckUserSubscriptionResponseData,
    )


__all__ = ("BroadcasterSubscription", "BroadcasterSubscriptions", "UserSubscription")


class UserSubscription:
    """Represents a subscription of a user.

    Attributes
    ----------
    broadcaster: PartialUser
        The broadcaster being subscribed to.
    gift: bool
        A Boolean value that determines whether the subscription is a gift subscription. Is ``True`` if the subscription was gifted.
    tier: typing.Literal["1000", "2000", "3000"]
        The type of subscription. Possible values are:

        - 1000: Tier 1
        - 2000: Tier 2
        - 3000: Tier 3

    gifter: PartialUser | None
        The user who gifted the subscription. This is None if ``gift`` is ``False``.
    """

    __slots__ = (
        "broadcaster",
        "gift",
        "gifter",
        "tier",
    )

    def __init__(
        self, data: BroadcasterSubscriptionsResponseData | CheckUserSubscriptionResponseData, *, http: HTTPClient
    ) -> None:
        self.broadcaster: PartialUser = PartialUser(
            data["broadcaster_id"], data["broadcaster_login"], data["broadcaster_name"], http=http
        )
        self.gift: bool = bool(data["is_gift"])
        self.tier: Literal["1000", "2000", "3000"] = data["tier"]
        _gifter_id, _gifter_login, _gifter_display_name = (
            data.get("gifter_id"),
            data.get("gifter_login"),
            data.get("gifter_name"),
        )
        self.gifter: PartialUser | None = (
            PartialUser(_gifter_id, _gifter_login, _gifter_display_name, http=http)
            if self.gift and _gifter_id is not None
            else None
        )

    def __repr__(self) -> str:
        return f"<UserSubscription broadcaster={self.broadcaster} tier={self.tier} gift={self.gift}>"

    @property
    def rounded_tier(self) -> int:
        """Returns the tier as a single digit. e.g. Tier 1000 = 1."""
        return round(int(self.tier) / 1000)


class BroadcasterSubscription(UserSubscription):
    """Represents a subscription of a user.

    Attributes
    ----------
    broadcaster: PartialUser
        The broadcaster being subscribed to.
    gift: bool
        A Boolean value that determines whether the subscription is a gift subscription. Is ``True`` if the subscription was gifted.
    tier: typing.Literal["1000", "2000", "3000"]
        The type of subscription. Possible values are:

        - 1000: Tier 1
        - 2000: Tier 2
        - 3000: Tier 3

    gifter: PartialUser | None
        The user who gifted the subscription. This is ``None`` if ``gift`` is ``False``.
    plan_name: str
        The name of the subscription.
    user: PartialUser
        The subscribing user.
    """

    __slots__ = ("plan_name", "user")

    def __init__(self, data: BroadcasterSubscriptionsResponseData, *, http: HTTPClient) -> None:
        super().__init__(data, http=http)
        self.plan_name: str = data["plan_name"]
        self.user: PartialUser = PartialUser(data["user_id"], data["user_login"], data["user_name"], http=http)

    def __repr__(self) -> str:
        return f"<BroadcasterSubscription broadcaster={self.broadcaster} tier={self.tier} gift={self.gift} plan_name={self.plan_name}>"


class BroadcasterSubscriptions:
    """Represents all the users that subscribe to a broadcaster.

    .. note::
        total and points are `None` if `user_ids` is provided.

    Attributes
    ----------
    subscriptions: HTTPAsyncIterator[BroadcasterSubscription]
        HTTPAsyncIterator of [`BroadcasterSubscription`][] objects.
    total: int | None
        The total number of users that subscribe to this broadcaster. This is `None` if `user_ids` is provided.
    points: int | None
        The current number of subscriber points earned by this broadcaster.
        Points are based on the subscription tier of each user that subscribes to this broadcaster.

        For example, a Tier 1 subscription is worth 1 point, Tier 2 is worth 2 points, and Tier 3 is worth 6 points.
        The number of points determines the number of emote slots that are unlocked for the broadcaster (see  `Subscriber Emote Slots <https://help.twitch.tv/s/article/subscriber-emote-guide#emoteslots>`_).

        This is `None` if `user_ids` is provided.
    """

    __slots__ = (
        "points",
        "subscriptions",
        "total",
    )

    def __init__(self, data: BroadcasterSubscriptionsResponse, iterator: HTTPAsyncIterator[BroadcasterSubscription]) -> None:
        self.subscriptions: HTTPAsyncIterator[BroadcasterSubscription] = iterator
        self.total: int | None = data.get("total")
        self.points: int | None = data.get("points")

    def __repr__(self) -> str:
        return f"<BroadcasterSubscriptions total={self.total} points={self.points}>"



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from twitchio.assets import Asset
from twitchio.user import PartialUser
from twitchio.utils import Colour, parse_timestamp


if TYPE_CHECKING:
    import datetime

    from twitchio.http import HTTPClient
    from twitchio.types_.responses import (
        BitsLeaderboardResponse,
        BitsLeaderboardResponseData,
        CheermotesResponseData,
        CheermotesResponseTiers,
        ExtensionTransactionsResponseCost,
        ExtensionTransactionsResponseData,
        ExtensionTransactionsResponseProductData,
    )


__all__ = (
    "BitLeaderboardUser",
    "BitsLeaderboard",
    "Cheermote",
    "CheermoteTier",
    "ExtensionCost",
    "ExtensionProductData",
    "ExtensionTransaction",
)


class BitsLeaderboard:
    """
    Represents a Bits leaderboard.

    Attributes
    ------------
    started_at: datetime.datetime | None
        The time the leaderboard started.
    ended_at: datetime.datetime | None
        The time the leaderboard ended.
    leaders: list[BitLeaderboardUser]
        The current leaders of the Leaderboard.
    """

    __slots__ = ("ended_at", "leaders", "started_at")

    def __init__(self, data: BitsLeaderboardResponse, *, http: HTTPClient) -> None:
        self.started_at = parse_timestamp(data["date_range"]["started_at"]) if data["date_range"]["started_at"] else None
        self.ended_at = parse_timestamp(data["date_range"]["ended_at"]) if data["date_range"]["ended_at"] else None
        self.leaders = [BitLeaderboardUser(d, http=http) for d in data["data"]]

    def __repr__(self) -> str:
        return f"<BitsLeaderboard started_at={self.started_at} ended_at={self.ended_at}>"


class BitLeaderboardUser:
    """A user's details within the Bit Leaderboard.

    Attributes
    ------------
    user: PartialUser
        The user.
    rank: int
        The user's position on the leaderboard.
    score: int
        The number of Bits the user has cheered.
    """

    __slots__ = ("rank", "score", "user")

    def __init__(self, data: BitsLeaderboardResponseData, *, http: HTTPClient) -> None:
        self.user: PartialUser = PartialUser(data["user_id"], data["user_login"], data["user_name"], http=http)
        self.rank: int = int(data["rank"])
        self.score: int = int(data["score"])

    def __repr__(self) -> str:
        return f"<BitLeaderboardUser user={self.user} rank={self.rank} score={self.score}>"


class CheermoteTier:
    """Represents a Cheermote tier.

    Attributes
    -----------
    min_bits: int
        The minimum bits for the tier
    id: str
        The ID of the tier
    colour: Colour
        The :class:`~twitchio.utils.Colour` of the tier. There is an alias named ``color``.
    images: dict[str, dict[str, dict[str, str]]]
        contains two dicts, ``light`` and ``dark``. Each item will have an ``animated`` and ``static`` item,
        which will contain yet another dict, with sizes ``1``, ``1.5``, ``2``, ``3``, and ``4``.
        Ex. ``cheermotetier.images["light"]["animated"]["1"]``
    can_cheer: bool
        Indicates whether emote information is accessible to users.
    show_in_bits_card: bool
        Indicates whether twitch hides the emote from the bits card.
    """

    __slots__ = ("_http", "can_cheer", "color", "colour", "id", "images", "min_bits", "show_in_bits_card")

    def __init__(self, data: CheermotesResponseTiers, *, http: HTTPClient) -> None:
        self._http: HTTPClient = http
        self.min_bits: int = data["min_bits"]
        self.id: str = data["id"]
        self.colour: Colour = Colour.from_hex(data["color"])
        self.color = self.colour
        self.images = data["images"]
        self.can_cheer: bool = data["can_cheer"]
        self.show_in_bits_card: bool = data["show_in_bits_card"]

    def __repr__(self) -> str:
        return f"<CheermoteTier id={self.id} min_bits={self.min_bits}>"

    def get_image(
        self,
        *,
        theme: Literal["light", "dark"] = "light",
        scale: Literal["1", "1.5", "2", "3", "4"] = "2",
        format: Literal["default", "static", "animated"] = "default",
    ) -> Asset | None:
        """Creates an :class:`~twitchio.Asset` for the cheermote, which can be used to download/save the cheermote image.

        Parameters
        ----------
        theme: typing.Literal["light", "dark"]
            The background theme of the cheermote. Defaults to "light".
        scale: str
            The scale (size) of the cheermote. Usually this will be one of: "1", "1.5", "2", "3", "4"
            Defaults to "2".
        format: typing.Literal["default", "static", "animated"]
            The format of the image for the cheermote. E.g a static image (PNG) or animated (GIF).

            Use "default" to get the default format for the emote, which will be animated if available, otherwise static.
            Defaults to "default".

        Examples
        --------
        .. code:: python3

            cheermotes: list[twitchio.Cheermote] = await client.fetch_cheermotes()
            cheermote: twitchio.Cheermote = cheermotes[0]

            # Get and save the emote asset as an image
            asset: twitchio.Asset = await cheeremote.tiers[0].get_image()
            await asset.save()

        Returns
        -------
        Asset | None
            The :class:`~twitchio.Asset` for the cheermote.
            You can use the asset to :meth:`~twitchio.Asset.read` or :meth:`~twitchio.Asset.save` the cheermote image or
            return the generated URL with :attr:`~twitchio.Asset.url`.
        """
        theme_images = self.images.get(theme, self.images.get("light", {}))
        format_images = theme_images.get("animated" if format in ("animated", "default") else "static", {})
        image_url = format_images.get(scale)

        if image_url is None:
            for format_type in ("animated", "static"):
                if theme_images.get(format_type):
                    image_url = next(iter(theme_images[format_type].values()), None)
                    if image_url:
                        break

        return None if image_url is None else Asset(image_url, http=self._http)


class Cheermote:
    """Represents a Cheermote

    +---------------------+------------------------------------------------------------------------------------------------------------------------+
    | Type                | Description                                                                                                            |
    +=====================+========================================================================================================================+
    | global_first_party  | A Twitch-defined Cheermote that is shown in the Bits card.                                                             |
    +---------------------+------------------------------------------------------------------------------------------------------------------------+
    | global_third_party  | A Twitch-defined Cheermote that is not shown in the Bits card.                                                         |
    +---------------------+------------------------------------------------------------------------------------------------------------------------+
    | channel_custom      | A broadcaster-defined Cheermote.                                                                                       |
    +---------------------+------------------------------------------------------------------------------------------------------------------------+
    | display_only        | Do not use; for internal use only.                                                                                     |
    +---------------------+------------------------------------------------------------------------------------------------------------------------+
    | sponsored           | A sponsor-defined Cheermote. When used, the sponsor adds additional Bits to the amount that the user cheered. For      |
    |                     | example, if the user cheered Terminator100, the broadcaster might receive 110 Bits, which includes the sponsor's 10    |
    |                     | Bits contribution.                                                                                                     |
    +---------------------+------------------------------------------------------------------------------------------------------------------------+

    Attributes
    -----------
    prefix: str
        The string used to Cheer that precedes the Bits amount.
    tiers: CheermoteTier
        The tiers this Cheermote has
    type: str
        Shows whether the emote is ``global_first_party``, ``global_third_party``, ``channel_custom``, ``display_only``, or ``sponsored``.
    order: int
        Order of the cheermotes as shown in the bits card, in ascending order.
    last_updated datetime.datetime
        The date this cheermote was last updated.
    charitable: bool
        Indicates whether this emote provides a charity contribution match during charity campaigns.
    """

    __slots__ = (
        "_http",
        "charitable",
        "last_updated",
        "order",
        "prefix",
        "tiers",
        "type",
    )

    def __init__(self, data: CheermotesResponseData, *, http: HTTPClient) -> None:
        self.prefix: str = data["prefix"]
        self.tiers = [CheermoteTier(d, http=http) for d in data["tiers"]]
        self.type: str = data["type"]
        self.order: int = int(data["order"])
        self.last_updated = parse_timestamp(data["last_updated"])
        self.charitable: bool = data["is_charitable"]

    def __repr__(self) -> str:
        return f"<Cheermote prefix={self.prefix} type={self.type} order={self.order}>"


class ExtensionTransaction:
    """Represents an Extension Transaction.

    Attributes
    -----------
    id: str
        An ID that identifies the transaction.
    timestamp: datetime.datetime
        The UTC date and time of the transaction.
    broadcaster: PartialUser
        The broadcaster that owns the channel where the transaction occurred.
    user: PartialUser
        The user that purchased the digital product.
    product_type: str
        The type of transaction. Currently only ``BITS_IN_EXTENSION``
    product_data: ExtensionProductData
        Details about the digital product.
    """

    __slots__ = ("broadcaster", "id", "product_data", "product_type", "timestamp", "user")

    def __init__(self, data: ExtensionTransactionsResponseData, *, http: HTTPClient) -> None:
        self.id: str = data["id"]
        self.timestamp: datetime.datetime = parse_timestamp(data["timestamp"])
        self.broadcaster = PartialUser(
            data["broadcaster_id"], data["broadcaster_login"], data["broadcaster_name"], http=http
        )
        self.user = PartialUser(data["user_id"], data["user_login"], http=http)
        self.product_type: str = data["product_type"]
        self.product_data: ExtensionProductData = ExtensionProductData(data["product_data"])

    def __repr__(self) -> str:
        return f"<ExtensionTransaction id={self.id} timestamp={self.timestamp} product_type={self.product_type}>"


class ExtensionProductData:
    """Represents Product Data of an Extension Transaction.

    Attributes
    -----------
    domain: str
        Set to twitch.ext. + <the extension's ID>.
    sku: str
        An ID that identifies the digital product.
    cost: ExtensionCost
        Contains details about the digital product's cost.
    in_development: bool
        Whether the product is in development.
    display_name: str
        The name of the digital product.
    expiration: str
        This field is always empty since you may purchase only unexpired products.
    broadcast: bool
        Whether the data was broadcast to all instances of the extension.
    """

    __slots__ = ("broadcast", "cost", "display_name", "domain", "expiration", "in_development", "sku")

    def __init__(self, data: ExtensionTransactionsResponseProductData) -> None:
        self.domain: str = data["domain"]
        self.sku: str = data["sku"]
        self.cost: ExtensionCost = ExtensionCost(data["cost"])
        self.in_development: bool = data["inDevelopment"]
        self.display_name: str = data["displayName"]
        self.expiration: str = data["expiration"]
        self.broadcast: bool = data["broadcast"]

    def __repr__(self) -> str:
        return f"<ExtensionProductData domain={self.domain} display_name={self.display_name} sku={self.sku}>"

    def __str__(self) -> str:
        return self.display_name


class ExtensionCost:
    """Represents Cost of an Extension Transaction.

    Attributes
    -----------
    amount: int
        The amount exchanged for the digital product.
    type: str
        The type of currency exchanged. Currently only ``bits``
    """

    __slots__ = ("amount", "type")

    def __init__(self, data: ExtensionTransactionsResponseCost) -> None:
        self.amount: int = int(data["amount"])
        self.type: str = data["type"]

    def __repr__(self) -> str:
        return f"<ExtensionCost amount={self.amount} type={self.type}>"

    def __str__(self) -> str:
        return str(self.amount)



"""
MIT License

Copyright (c) 2017 - Present PythonistaGuild

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import datetime
import logging
from collections.abc import Callable, Coroutine
from typing import Any, TypeAlias, TypeVar

from twitchio.backoff import Backoff


__all__ = ("Routine", "routine")


T = TypeVar("T")
CoroT: TypeAlias = Callable[..., Coroutine[Any, Any, Any]]


LOGGER: logging.Logger = logging.getLogger(__name__)


def compute_timedelta(dt: datetime.datetime) -> float:
    if dt.tzinfo is None:
        dt = dt.astimezone()

    now = datetime.datetime.now(tz=datetime.UTC)
    return max((dt - now).total_seconds(), 0)


class Routine:
    """The TwitchIO Routine class which runs asynchronously in the background on a timed loop.

    .. note::

        You should not instantiate this class manually, instead use the :func:`routine` decorator instead.


    Examples
    --------

    .. code:: python3

        # This routine will run every minute until stopped or canceled.

        @routines.routine(delta=datetime.timedelta(minutes=1))
        async def my_routine() -> None:
            print("Hello World!")

        my_routine.start()

    .. code:: python3

        # Pass some arguments to a routine...

        @routines.routine(delta=datetime.timedelta(minutes=1))
        async def my_routine(hello: str) -> None:
            print(f"Hello {hello}")

        my_routine.start("World!")

    .. code:: python3

        # Only run the routine three of times...

        @routines.routine(delta=datetime.timedelta(minutes=1), iterations=3)
        async def my_routine(hello: str) -> None:
            print(f"Hello {hello}")

        my_routine.start("World!")

    """

    def __init__(
        self,
        *,
        name: str | None = None,
        coro: CoroT,
        max_attempts: int | None = None,
        iterations: int | None,
        wait_remainder: bool = False,
        stop_on_error: bool = False,
        wait_first: bool,
        time: datetime.datetime | None,
        delta: datetime.timedelta | None,
    ) -> None:
        self._coro: CoroT = coro
        self._name: str = name or f"twitchio.ext.routines: <{self.__class__.__qualname__}[{self._coro.__qualname__}]>"
        self._task: asyncio.Task[None] | None = None
        self._injected = None
        self._time: datetime.datetime | None = time
        self._original_delta: datetime.timedelta | None = delta
        self._delta: float | None = delta.total_seconds() if delta else None

        self._before_routine: CoroT | None = None
        self._after_routine: CoroT | None = None
        self._on_error: CoroT | None = None

        self._stop_on_error: bool = stop_on_error
        self._should_stop: bool = False
        self._restarting: bool = False
        self._wait_first: bool = wait_first

        self._completed: int = 0
        self._iterations: int | None = iterations
        self._current_iteration: int = 0
        self._wait_remainder: bool = wait_remainder
        self._last_start: datetime.datetime | None = None

        self._max_attempts: int | None = max_attempts

        self._args: Any = ()
        self._kwargs: Any = {}

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}[{self._coro.__qualname__}]>"

    def __get__(self, instance: T, type_: type[T]) -> Routine:
        if instance is None:
            return self

        copy: Routine = Routine(
            coro=self._coro,
            name=self._name,
            time=self._time,
            delta=self._original_delta,
            max_attempts=self._max_attempts,
            iterations=self._iterations,
            wait_remainder=self._wait_remainder,
            wait_first=self._wait_first,
            stop_on_error=self._stop_on_error,
        )
        copy._injected = instance
        copy._before_routine = self._before_routine
        copy._after_routine = self._after_routine
        copy._on_error = self._on_error

        setattr(instance, self._coro.__name__, copy)
        return copy

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self._injected:
            args = (self._injected, *args)

        await self._coro(*args, **kwargs)

    def _can_cancel(self) -> bool:
        return self._task is not None and not self._task.done()

    async def _call_error(self, error: Exception) -> None:
        if self._on_error is None:
            await self.on_error(error)
            return

        if self._injected is not None:
            await self._on_error(self._injected, error)
        else:
            await self._on_error(error)

    async def _routine_loop(self, *args: Any, **kwargs: Any) -> None:
        backoff: Backoff = Backoff(base=3, maximum_time=10, maximum_tries=5)

        if self._should_stop:
            return

        if self._before_routine:
            try:
                await self._before_routine(*args, **kwargs)
            except Exception as e:
                await self._call_error(e)
                return self.cancel()

        if self._wait_first:
            if self._time:
                await asyncio.sleep(compute_timedelta(self._time))
            elif self._delta:
                await asyncio.sleep(self._delta)

        attempts: int | None = self._max_attempts
        try:
            while True:
                self._last_start: datetime.datetime | None = datetime.datetime.now(tz=datetime.UTC)
                self._current_iteration += 1

                try:
                    await self._coro(*args, **kwargs)
                except Exception as e:
                    await self._call_error(e)

                    if self._stop_on_error:
                        self._should_stop = False
                        break

                    if attempts is not None:
                        attempts -= 1

                        if attempts <= 0:
                            LOGGER.warning(
                                "The maximum retry attempts for Routine: %s has been reached and will now be canceled.",
                                self.__repr__(),
                            )
                            break

                        await asyncio.sleep(backoff.calculate())
                        continue

                else:
                    attempts = self._max_attempts
                    self._completed += 1

                if self.remaining_iterations is not None and self.remaining_iterations <= 0:
                    break

                if self._should_stop:
                    self._should_stop = False
                    break

                if self._time:
                    sleep = compute_timedelta(self._time + datetime.timedelta(days=self._current_iteration))
                else:
                    assert self._delta is not None

                    if not self._wait_remainder:
                        sleep = self._delta
                    else:
                        maxxed = (self._last_start - datetime.datetime.now(tz=datetime.UTC)).total_seconds()
                        sleep = max(maxxed + self._delta, 0)

                await asyncio.sleep(sleep)

        except Exception as e:
            msg = "A fatal error occured during an iteration of Routine: %s. This routine will now be canceled."
            LOGGER.error(msg, self.__repr__(), exc_info=e)

        if self._after_routine:
            try:
                await self._after_routine(*args, **kwargs)
            except Exception as e:
                await self._call_error(e)

        self.cancel()

    @property
    def args(self) -> tuple[Any, ...]:
        """Property returning any positional arguments passed to the routine via :meth:`.start`."""
        return self._args

    @property
    def kwargs(self) -> Any:
        """Property returning any keyword arguments passed to the routine via :meth:`.start`."""
        return self._kwargs

    def start(self, *args: Any, **kwargs: Any) -> asyncio.Task[None]:
        r"""Method to start the :class:`~Routine` in the background.

        .. note::

            You can not start an already running task. See: :meth:`.restart` instead.

        Parameters
        ----------
        *args: Any
            Any positional args passed to this method will also be passed to your :class:`~Routine` callback.
        **kwargs: Any
            Any keyword arguments passed to this method will also be passed to your :class:`~Routine` callback.

        Returns
        -------
        :class:`asyncio.Task`
            The internal background task associated with this :class:`~Routine`.
        """
        if self._task and not self._task.done() and not self._restarting:
            raise RuntimeError(f"Routine {self!r} is currently running and has not completed.")

        self._args = args
        self._kwargs = kwargs

        if self._injected:
            args = (self._injected, *args)

        self._restarting = False

        loop = self._routine_loop(*args, **kwargs)
        self._task = asyncio.create_task(loop, name=self._name)

        return self._task

    def stop(self) -> None:
        """Method to stop the currently running task after it completes its current iteration.

        Unlike :meth:`.cancel` this will schedule the task to stop after it completes its next iteration.
        """
        self._should_stop = True

    def cancel(self) -> None:
        """Method to immediately cancel the currently running :class:`~Routine`.

        Unlike :meth:`.stop`, this method is not graceful and will cancel the task in its current iteration.
        """
        if self._can_cancel():
            assert self._task is not None
            self._task.cancel()

        if not self._restarting:
            self._task = None

    def restart(self, *, force: bool = True) -> None:
        """Method which restarts the :class:`~Routine`.

        If the :class:`~Routine` has not yet been started, this method immediately returns.

        Parameters
        ----------
        force: bool
            Whether to cancel the currently running routine and restart it immediately. If this is ``False`` the
            :class:`~Routine` will start after it's current iteration. Defaults to ``True``.
        """
        self._restarting = True
        self._current_iteration = 0

        if not self._task:
            return

        def restart_when_over(fut: asyncio.Task[None]) -> None:
            fut.remove_done_callback(restart_when_over)
            self.start(*self._args, **self._kwargs)

        if self._can_cancel():
            self._task.add_done_callback(restart_when_over)

            if force:
                self._task.cancel()
            else:
                self.stop()

    def before_routine(self, func: CoroT) -> None:
        """|deco|

        Decorator used to set a coroutine to run before the :class:`~Routine` has started.

        Any arguments passed to :meth:`.start` will also be passed to this coroutine callback.
        """
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f'"before_routine" for {self!r} expected a coroutine function not {type(func).__name__!r}')

        if self._before_routine is not None:
            LOGGER.warning("The before_routine for %s has previously been set.", self.__repr__())

        self._before_routine = func

    def after_routine(self, func: CoroT) -> None:
        """|deco|

        Decorator used to set a coroutine to run after the :class:`~Routine` has been stopped, canceled or completed.

        Any arguments passed to :meth:`.start` will also be passed to this coroutine callback.
        """
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f'"after_routine" for {self!r} expected a coroutine function not {type(func).__name__!r}')

        if self._after_routine is not None:
            LOGGER.warning("The after_routine for %s has previously been set.", self.__repr__())

        self._after_routine = func

    async def on_error(self, error: Exception) -> None:
        """|coro|

        Default error handler for this :class:`~Routine`. You can override this with the :meth:`.error` decorator.

        By default this handler will log the exception.

        Parameters
        ----------
        error: Exception
            The exception raised in the routine.
        """
        msg = "Ignoring Exception in Routine %s: %s\n"
        LOGGER.error(msg, self.__repr__(), error, exc_info=error)

    def error(self, func: CoroT) -> None:
        """|deco|

        Decorator used to set a coroutine as an error handler for this :class:`~Routine`.
        """
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f'"error" for {self!r} expected a coroutine function not {type(func).__name__!r}')

        self._on_error = func

    @property
    def completed_iterations(self) -> int:
        """Property returning a :class:`int` of the successfully completed iterations.

        .. note::

            Iterations where an error occured do not count towards completed iterations.
        """
        return self._completed

    @property
    def remaining_iterations(self) -> int | None:
        """Property returning a :class:`int` or :class:`None` of the remaining iterations.

        If ``iterations`` was not set this returns ``None``.
        """
        return None if self._iterations is None else self._iterations - self._completed

    @property
    def current_iteration(self) -> int:
        """Property returning the current iteration count as a :class:`int`."""
        return self._current_iteration

    @property
    def last_iteration(self) -> datetime.datetime | None:
        """Property which returns a :class:`datetime.datetime` of when the current iteration started."""
        return self._last_start

    def next_iteration(self) -> float:
        """Method which returns the time until the next iteration as a float of seconds."""
        if self._time:
            return compute_timedelta(self._time + datetime.timedelta(days=self._current_iteration))

        assert self._delta is not None

        if self._last_start is None:
            return self._delta
        else:
            return max((self._last_start - datetime.datetime.now(tz=datetime.UTC)).total_seconds() + self._delta, 0)


def routine(
    *,
    delta: datetime.timedelta | None = None,
    time: datetime.datetime | None = None,
    name: str | None = None,
    iterations: int | None = None,
    wait_first: bool = False,
    wait_remainder: bool = False,
    max_attempts: int | None = 5,
    stop_on_error: bool = False,
) -> Callable[[CoroT], Routine]:
    """|deco|

    A decorator to assign a coroutine as a :class:`~Routine`.

    .. important::

        One parameter of either ``time`` or ``delta`` must be passed. Both parameters CAN NOT be used together.

    Parameters
    ----------
    delta: :class:`datetime.timedelta` | None
        A :class:`datetime.timedelta` of time to wait per iteration of the :class:`~Routine`. If this is ``None``,
        you must pass the ``time`` parameter. Defaults to ``None``.
    time: :class:`datetime.datetime` | None
        A :class:`datetime.datetime` to schedule an run each iteration of the :class:`~Routine`. The :class:`~Routine` will
        run at the same time everyday. If this is ``None``, you must pass the ``delta`` parameter. Defaults to ``None``.
    name: str | None
        An optional name to use for this routine. Defaults to ``None`` which uses the repr of the :class:`~Routine`.
    iterations: int | None
        An optional :class:`int` amount of iterations to run this routine for. For example if set to ``3``, the
        :class:`~Routine` will only run ``3`` times and complete. If set to ``None`` the routine will continue running
        until stopped or canceled. Defaults to ``None``.
    wait_first: bool
        Whether the :class:`~Routine` should wait the amount of time specified in ``time`` or ``delta`` before starting.
        Defaults to ``False``, which will run the first iteration immediately after :meth:`~Routine.start` is called.
    wait_remainder: bool
        Whether to wait **only** the remaining time to the next iteration after the current iteration completes.
        For example if your :class:`~Routine` is scheduled to run every minute, and your iteration took ``30 seconds``,
        the routine will only wait an additional ``30 seconds`` before running its next iteration. Defaults to ``False``
        which means that the :class:`~Routine` will wait the full time **after** each iteration. Note: this has no effect
        when the ``time`` parameter is passed.
    max_attempts: int | None
        The maximum times the :class:`~Routine` should be allowed to error consecuitevely before stopping. When an error
        occurs and this parameter is set, the routine will sleep for a small backoff interval before retrying the iteration.
        If the ``max_attempts`` counter is hit after multiple consecutive errors, the routine will stop. This counter resets
        when a successful iteration occurs. Note: ``max_attempts`` has no effect when ``stop_on_error`` is set.
        If set to ``None``, the routine will backoff indefinitely. Defaults to ``5``.
    stop_on_error: bool
        A bool indicating whether the :class:`~Routine` should immediately stop when encountering an error. This parameter
        takes precedence over ``max_attempts``. Defaults to ``False``.

    Raises
    ------
    :class:`TypeError`
        The function decorated is not a coroutine function.
    :class:`RuntimeError`
        The ``time`` and ``delta`` arguments were both passed together. Only one of these parameters can be used.
    :class:`RuntimeError`
        Both the ``time`` and ``delta`` arguments are missing. One of these parameters must be used.
    """

    def wrapper(func: CoroT) -> Routine:
        nonlocal time

        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f"Routine expected coroutine function not {type(func).__name__!r}")

        if not time and not delta:
            raise RuntimeError('One of either the "time" or "delta" arguments must be passed.')

        if time is not None and delta is not None:
            raise RuntimeError(
                'The "time" argument can not be used in conjunction with the "delta" argument. Only one should be set.'
            )

        if not time:
            delta_ = delta
        else:
            delta_ = None

            now = datetime.datetime.now(time.tzinfo)
            if time < now:
                time = datetime.datetime.combine(now.date(), time.time())
            if time < now:
                time = time + datetime.timedelta(days=1)

        return Routine(
            coro=func,
            time=time,
            delta=delta_,
            name=name,
            iterations=iterations,
            wait_first=wait_first,
            wait_remainder=wait_remainder,
            max_attempts=max_attempts,
            stop_on_error=stop_on_error,
        )

    return wrapper
